/*-------------------------------------------------------------------------+
| Copyright 2022 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.variability.util;

import static org.fortiss.tooling.ext.variability.model.VariabilityModelElementFactory.createOptionalVariationPointSpecification;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getFirstChildWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.ext.variability.model.DeactivationSpecification;
import org.fortiss.tooling.ext.variability.model.OptionalVariationPointSpecification;
import org.fortiss.tooling.kernel.service.ICommandStackService;
import org.fortiss.variability.model.features.AbstractFeatureModel;
import org.fortiss.variability.model.presence.PresenceCondition;
import org.fortiss.variability.presence.compiler.PresenceConditionCompiler;

/**
 * Utility functions related to variability in AF3.
 * 
 * @author bayha
 */
public class VariabilityUtils {

	/**
	 * Checks for the given {@link EObject} whether it has been deactivated by a
	 * {@link DeactivationSpecification}.
	 * 
	 * @param eObj
	 *            {@link EObject} to be checked for deactivation.
	 * 
	 * @return 'true' if eObj contains a {@link DeactivationSpecification}, 'false'
	 *         otherwise.
	 */
	public static boolean isDeactivated(EObject eObj) {
		if(!(eObj instanceof IModelElement)) {
			return true;
		}

		IModelElement element = (IModelElement)eObj;

		DeactivationSpecification dSpec =
				pickFirstInstanceOf(DeactivationSpecification.class, element.getSpecifications());
		return dSpec != null;
	}

	/**
	 * Checks for the given {@link EObject} whether it can contain any variability.
	 * 
	 * Note, that currently this method checks for the existence of an {@link AbstractFeatureModel}
	 * in the project, only.
	 * 
	 * @param element
	 *            The {@link EObject} to check for possible variability.
	 * @return 'true' if the given element can contain any variability.
	 */
	public static boolean canContainVariability(EObject element) {
		if(element instanceof AbstractFeatureModel) {
			return true;
		}

		// Find the root container (usually a FileProject)
		EObject topLevelElem = element;
		while(topLevelElem.eContainer() != null) {
			topLevelElem = topLevelElem.eContainer();
		}

		// Check whether this root container contains any AbstractFeatureModel.
		return getFirstChildWithType(topLevelElem, AbstractFeatureModel.class) != null;
	}

	/**
	 * Checks whether the given {@link IModelElement} contains a non-empty presence
	 * condition. I.e. whether it is actually really optional.
	 * 
	 * @param elem
	 *            The {@link IModelElement} to be checked.
	 * 
	 * @return 'true' if elem contains a {@link PresenceCondition}, 'false'
	 *         otherwise.
	 */
	public static boolean hasPresenceCondition(IModelElement elem) {
		return getPresenceCondition(elem) != null;
	}

	/**
	 * Retrieves the {@link PresenceCondition} from the given {@link IModelElement}.
	 * 
	 * @param elem
	 *            The {@link IModelElement} to get the {@link PresenceCondition}
	 *            from.
	 * 
	 * @return The {@link PresenceCondition} contained in elems
	 *         {@link OptionalVariationPointSpecification}. 'null' if it contains
	 *         none.
	 */
	public static PresenceCondition getPresenceCondition(IModelElement elem) {
		OptionalVariationPointSpecification opt = getOptVarPointSpecification(elem);

		if(opt != null) {
			return opt.getPresenceCondition();
		}

		return null;
	}

	/**
	 * Retrieves the {@link OptionalVariationPointSpecification} from the given
	 * {@link IModelElement}.
	 * 
	 * @param elem
	 *            The {@link IModelElement} to get the
	 *            {@link OptionalVariationPointSpecification} from.
	 * 
	 * @return The {@link OptionalVariationPointSpecification} of elem. 'null' is
	 *         there should not be any.
	 */
	public static OptionalVariationPointSpecification
			getOptVarPointSpecification(IModelElement elem) {
		OptionalVariationPointSpecification opt = pickFirstInstanceOf(
				OptionalVariationPointSpecification.class, elem.getSpecifications());

		return opt;
	}

	/**
	 * Compiles and sets the given presence condition String for the given
	 * {@link IModelElement}.
	 * 
	 * @param elem
	 *            The {@link IModelElement} to set the {@link PresenceCondition}
	 *            for.
	 * @param pcStr
	 *            The presence condition {@link String} to be compiled to a
	 *            {@link PresenceCondition} and set for elem.
	 * 
	 * @return 'true' if the {@link String} could be parsed correctly and set for
	 *         the {@link IModelElement}. 'false' otherwise.
	 */
	public static boolean setPresenceCondition(IModelElement elem, String pcStr) {
		PresenceConditionCompiler comp = new PresenceConditionCompiler();
		PresenceCondition pc = comp.compilePresenceCondition(elem, pcStr);

		if(pc == null) {
			if(comp.errors.size() > 0) {
				return false;
			}
		}

		OptionalVariationPointSpecification spec = getOptVarPointSpecification(elem);

		if(spec == null) {
			if(pc == null) {
				// No existing PC. No PC given. Nothing to do.
				return true;
			}

			spec = createOptionalVariationPointSpecification();
			final OptionalVariationPointSpecification finalSpec = spec;
			ICommandStackService.getInstance().runAsCommand(elem,
					() -> elem.addSpecification(finalSpec));
		}

		final OptionalVariationPointSpecification finalSpec = spec;
		ICommandStackService.getInstance().runAsCommand(elem,
				() -> finalSpec.setPresenceCondition(pc));

		return true;
	}
}
